"""Module for get all node information from gateway."""
from enum import Enum
from datetime import datetime
import struct

from pyvlx.const import Command, NodeTypeWithSubtype, NodeVariation, Velocity
from pyvlx.position import Position
from pyvlx.alias_array import AliasArray
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase


class FrameGetAllNodesInformationRequest(FrameBase):
    """Frame for get node information request."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_REQ)


class AllNodesInformationStatus(Enum):
    """Enum for node information status."""

    OK = 0  # pylint: disable=invalid-name
    Error_System_Table_Empty = 1


class FrameGetAllNodesInformationConfirmation(FrameBase):
    """Frame for confirmation for node information request."""

    PAYLOAD_LEN = 2

    def __init__(self, status=AllNodesInformationStatus.OK, number_of_nodes=0):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_CFM)
        self.status = status
        self.number_of_nodes = number_of_nodes

    def get_payload(self):
        """Return Payload."""
        return bytes([self.status.value, self.number_of_nodes])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = AllNodesInformationStatus(payload[0])
        self.number_of_nodes = payload[1]

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetAllNodesInformationConfirmation status=\'{}\' number_of_nodes={}/>'.format(self.status, self.number_of_nodes)


class FrameGetAllNodesInformationNotification(FrameBase):
    """Frame for notification of all nodes information request."""

    PAYLOAD_LEN = 124

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_NTF)
        self.node_id = 0
        self.order = 0
        self.placement = 0
        self.name = ""
        self.velocity = Velocity.DEFAULT
        self.node_type = NodeTypeWithSubtype.NO_TYPE
        self.product_group = 0
        self.product_type = 0
        self.node_variation = NodeVariation.NOT_SET
        self.power_mode = 0
        self.build_number = 0
        self._serial_number = bytes(8)
        self.state = 0
        self.current_position = Position()
        self.target = Position()
        self.current_position_fp1 = Position()
        self.current_position_fp2 = Position()
        self.current_position_fp3 = Position()
        self.current_position_fp4 = Position()
        self.remaining_time = 0
        self.timestamp = 0
        self.alias_array = AliasArray()

    @property
    def serial_number(self):
        """Property for serial number in a human readable way."""
        return ":".join("{:02x}".format(c) for c in self._serial_number)

    def get_payload(self):
        """Return Payload."""
        payload = bytes()
        payload += bytes([self.node_id])
        payload += bytes([self.order >> 8 & 255, self.order & 255])
        payload += bytes([self.placement])
        payload += bytes(string_to_bytes(self.name, 64))
        payload += bytes([self.velocity.value])
        payload += bytes([self.node_type.value >> 8 & 255, self.node_type.value & 255])
        payload += bytes([self.product_group])
        payload += bytes([self.product_type])
        payload += bytes([self.node_variation.value])
        payload += bytes([self.power_mode])
        payload += bytes([self.build_number])
        payload += bytes(self._serial_number)
        payload += bytes([self.state])
        payload += bytes(self.current_position.raw)
        payload += bytes(self.target.raw)
        payload += bytes(self.current_position_fp1.raw)
        payload += bytes(self.current_position_fp2.raw)
        payload += bytes(self.current_position_fp3.raw)
        payload += bytes(self.current_position_fp4.raw)
        payload += bytes([self.remaining_time >> 8 & 255, self.remaining_time & 255])
        payload += struct.pack(">I", self.timestamp)
        payload += bytes(self.alias_array)

        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_id = payload[0]
        self.order = payload[1] * 256 + payload[2]
        self.placement = payload[3]
        self.name = bytes_to_string(payload[4:68])
        self.velocity = Velocity(payload[68])
        self.node_type = NodeTypeWithSubtype(payload[69] * 256 + payload[70])
        self.product_group = payload[71]
        self.product_type = payload[72]
        self.node_variation = NodeVariation(payload[73])
        self.power_mode = payload[74]
        self.build_number = payload[75]
        self._serial_number = payload[76:84]
        self.state = payload[84]
        self.current_position = Position(payload[85:87])
        self.target = Position(payload[87:89])
        self.current_position_fp1 = Position(payload[89:91])
        self.current_position_fp2 = Position(payload[91:93])
        self.current_position_fp3 = Position(payload[93:95])
        self.current_position_fp4 = Position(payload[95:97])
        self.remaining_time = payload[97] * 256 + payload[98]
        self.timestamp = struct.unpack(">I", payload[99:103])[0]
        self.alias_array = AliasArray(payload[103:125])

    @property
    def timestamp_formatted(self):
        """Return time as human readable string."""
        return datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S')

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetAllNodesInformationNotification node_id={} order={} ' \
            'placement={} name=\'{}\' velocity={} node_type=\'{}\' product_group={} ' \
            'product_type={} node_variation={} power_mode={} build_number={} ' \
            'serial_number=\'{}\' state={} current_position=\'{}\' ' \
            'target=\'{}\' current_position_fp1=\'{}\' current_position_fp2=\'{}\' ' \
            'current_position_fp3=\'{}\' current_position_fp4=\'{}\' ' \
            'remaining_time={} time=\'{}\' alias_array=\'{}\'/>'.format(
                self.node_id,
                self.order,
                self.placement,
                self.name,
                self.velocity,
                self.node_type,
                self.product_group,
                self.product_type,
                self.node_variation,
                self.power_mode,
                self.build_number,
                self.serial_number,
                self.state,
                self.current_position,
                self.target,
                self.current_position_fp1,
                self.current_position_fp2,
                self.current_position_fp3,
                self.current_position_fp4,
                self.remaining_time,
                self.timestamp_formatted,
                self.alias_array)


class FrameGetAllNodesInformationFinishedNotification(FrameBase):
    """Frame for notification of all nodes information finished notification."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_ALL_NODES_INFORMATION_FINISHED_NTF)



"""Module for all KLF 200 API frames."""

# flake8: noqa
from .frame import FrameBase
from .frame_helper import calc_crc, extract_from_frame
from .frame_get_scene_list import FrameGetSceneListRequest, FrameGetSceneListConfirmation, FrameGetSceneListNotification
from .frame_get_node_information import FrameGetNodeInformationRequest, FrameGetNodeInformationConfirmation, FrameGetNodeInformationNotification
from .frame_get_all_nodes_information import FrameGetAllNodesInformationRequest, FrameGetAllNodesInformationConfirmation, \
    FrameGetAllNodesInformationNotification, FrameGetAllNodesInformationFinishedNotification
from .frame_password_enter import FramePasswordEnterRequest, FramePasswordEnterConfirmation, PasswordEnterConfirmationStatus
from .frame_discover_nodes import FrameDiscoverNodesRequest, FrameDiscoverNodesConfirmation, FrameDiscoverNodesNotification
from .frame_error_notification import FrameErrorNotification, ErrorType
from .frame_command_send import FrameCommandSendRequest, FrameCommandSendConfirmation, FrameCommandRunStatusNotification, \
    FrameCommandRemainingTimeNotification, FrameSessionFinishedNotification, CommandSendConfirmationStatus
from .frame_activate_scene import FrameActivateSceneRequest, FrameActivateSceneConfirmation, ActivateSceneConfirmationStatus
from .frame_get_protocol_version import FrameGetProtocolVersionRequest, FrameGetProtocolVersionConfirmation
from .frame_get_version import FrameGetVersionRequest, FrameGetVersionConfirmation
from .frame_set_node_name import FrameSetNodeNameRequest, FrameSetNodeNameConfirmation, SetNodeNameConfirmationStatus
from .frame_node_information_changed import FrameNodeInformationChangedNotification
from .frame_get_state import FrameGetStateRequest, FrameGetStateConfirmation, GatewayState, GatewaySubState
from .frame_activation_log_updated import FrameActivationLogUpdatedNotification
from .frame_house_status_monitor_disable_cfm import FrameHouseStatusMonitorDisableConfirmation
from .frame_house_status_monitor_disable_req import FrameHouseStatusMonitorDisableRequest
from .frame_house_status_monitor_enable_cfm import FrameHouseStatusMonitorEnableConfirmation
from .frame_house_status_monitor_enable_req import FrameHouseStatusMonitorEnableRequest
from .frame_set_utc_req import FrameSetUTCRequest
from .frame_set_utc_cfm import FrameSetUTCConfirmation
from .frame_node_state_position_changed_notification import FrameNodeStatePositionChangedNotification



"""Module for get node information from gateway."""
from datetime import datetime
import struct

from pyvlx.const import Command
from pyvlx.position import Position

from .frame import FrameBase


class FrameNodeStatePositionChangedNotification(FrameBase):
    """Frame for notification of note information request."""

    PAYLOAD_LEN = 20

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_NODE_STATE_POSITION_CHANGED_NTF)
        self.node_id = 0
        self.state = 0
        self.current_position = Position()
        self.target = Position()
        self.current_position_fp1 = Position()
        self.current_position_fp2 = Position()
        self.current_position_fp3 = Position()
        self.current_position_fp4 = Position()
        self.remaining_time = 0
        self.timestamp = 0

    def get_payload(self):
        """Return Payload."""
        payload = bytes([self.node_id])
        payload += bytes([self.state])
        payload += bytes(self.current_position.raw)
        payload += bytes(self.target.raw)
        payload += bytes(self.current_position_fp1.raw)
        payload += bytes(self.current_position_fp2.raw)
        payload += bytes(self.current_position_fp3.raw)
        payload += bytes(self.current_position_fp4.raw)
        payload += bytes([self.remaining_time >> 8 & 255, self.remaining_time & 255])
        payload += struct.pack(">I", self.timestamp)
        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_id = payload[0]
        self.state = payload[1]
        self.current_position = Position(payload[2:4])
        self.target = Position(payload[4:6])
        self.current_position_fp1 = Position(payload[6:8])
        self.current_position_fp2 = Position(payload[8:10])
        self.current_position_fp3 = Position(payload[10:12])
        self.current_position_fp4 = Position(payload[12:14])
        self.remaining_time = payload[14] * 256 + payload[15]
        # @VELUX: looks like your timestamp is wrong. Looks like
        # you are only transmitting the two lower bytes.
        self.timestamp = struct.unpack(">I", payload[16:20])[0]

    @property
    def timestamp_formatted(self):
        """Return time as human readable string."""
        return datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S')

    def __str__(self):
        """Return human readable string."""
        return '<FrameNodeStatePositionChangedNotification node_id={} ' \
            'state={} current_position=\'{}\' ' \
            'target=\'{}\' current_position_fp1=\'{}\' current_position_fp2=\'{}\' ' \
            'current_position_fp3=\'{}\' current_position_fp4=\'{}\' ' \
            'remaining_time={} time=\'{}\'/>'.format(
                self.node_id,
                self.state,
                self.current_position,
                self.target,
                self.current_position_fp1,
                self.current_position_fp2,
                self.current_position_fp3,
                self.current_position_fp4,
                self.remaining_time,
                self.timestamp_formatted)



"""Module for Frames."""
import struct

from pyvlx.exception import PyVLXException

from .frame_helper import calc_crc


class FrameBase:
    """Class for Base Frame."""

    def __init__(self, command):
        """Initialize Base Frame."""
        self.command = command

    def __bytes__(self):
        """Get raw bytes of Frame."""
        payload = self.get_payload()
        self.validate_payload_len(payload)
        return self.build_frame(self.command, payload)

    def validate_payload_len(self, payload):
        """Validate payload len."""
        if not hasattr(self, "PAYLOAD_LEN"):
            # No fixed payload len, e.g. within FrameGetSceneListNotification
            return
        # pylint: disable=no-member
        if len(payload) != self.PAYLOAD_LEN:
            raise PyVLXException("Invalid payload len", expected_len=self.PAYLOAD_LEN, current_len=len(payload), frame_type=type(self).__name__)

    def get_payload(self):
        """Return Payload."""
        # pylint: disable=no-self-use
        return b''

    def from_payload(self, payload):
        """Init frame from binary data."""
        # pylint: disable=no-self-use

    def __str__(self):
        """Return human readable string."""
        return '<{}/>'.format(type(self).__name__)

    @staticmethod
    def build_frame(command, payload):
        """Build raw bytes from command and payload."""
        packet_length = 2 + len(payload) + 1
        ret = struct.pack("BB", 0, packet_length)
        ret += struct.pack(">H", command.value)
        ret += payload
        ret += struct.pack("B", calc_crc(ret))
        return ret



"""Module for password enter frame classes."""
from enum import Enum

from pyvlx.const import Command
from pyvlx.exception import PyVLXException
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase


class FramePasswordEnterRequest(FrameBase):
    """Frame for sending password enter request."""

    MAX_SIZE = 32
    PAYLOAD_LEN = 32

    def __init__(self, password=None):
        """Init Frame."""
        super().__init__(Command.GW_PASSWORD_ENTER_REQ)
        self.password = password

    def get_payload(self):
        """Return Payload."""
        if self.password is None:
            raise PyVLXException("password is none")
        if len(self.password) > self.MAX_SIZE:
            raise PyVLXException("password is too long")
        return string_to_bytes(self.password, self.MAX_SIZE)

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.password = bytes_to_string(payload)

    def __str__(self):
        """Return human readable string."""
        password_esc = None if self.password is None else '{}****'.format(self.password[:2])
        return '<FramePasswordEnterRequest password={}/>'.format(password_esc)


class PasswordEnterConfirmationStatus(Enum):
    """Enum class for status of password enter confirmation."""

    SUCCESSFUL = 0
    FAILED = 1


class FramePasswordEnterConfirmation(FrameBase):
    """Frame for confirmation for sent password."""

    PAYLOAD_LEN = 1

    def __init__(self, status=PasswordEnterConfirmationStatus.SUCCESSFUL):
        """Init Frame."""
        super().__init__(Command.GW_PASSWORD_ENTER_CFM)
        self.status = status

    def get_payload(self):
        """Return Payload."""
        return bytes([self.status.value])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = PasswordEnterConfirmationStatus(payload[0])

    def __str__(self):
        """Return human readable string."""
        return '<FramePasswordEnterConfirmation status=\'{}\'/>'.format(self.status)



"""Module for house status monitor."""
from .api_event import ApiEvent
from .exception import PyVLXException
from .frames import (FrameHouseStatusMonitorDisableConfirmation,
                     FrameHouseStatusMonitorDisableRequest,
                     FrameHouseStatusMonitorEnableConfirmation,
                     FrameHouseStatusMonitorEnableRequest)


class HouseStatusMonitorEnable(ApiEvent):
    """Class for enabling house status monotor."""

    def __init__(self, pyvlx):
        """Initialize login class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameHouseStatusMonitorEnableConfirmation):
            return False
        self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameHouseStatusMonitorEnableRequest()


class HouseStatusMonitorDisable(ApiEvent):
    """Class for disabling house status monotor."""

    def __init__(self, pyvlx):
        """Initialize login class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameHouseStatusMonitorDisableConfirmation):
            return False
        self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameHouseStatusMonitorDisableRequest()


async def house_status_monitor_enable(pyvlx):
    """Enable house status monitor."""
    status_monitor_enable = HouseStatusMonitorEnable(pyvlx=pyvlx)
    await status_monitor_enable.do_api_call()
    if not status_monitor_enable.success:
        raise PyVLXException("Unable enable house status monitor.")


async def house_status_monitor_disable(pyvlx):
    """Disable house status monitor."""
    status_monitor_disable = HouseStatusMonitorDisable(pyvlx=pyvlx)
    await status_monitor_disable.do_api_call()
    if not status_monitor_disable.success:
        raise PyVLXException("Unable disable house status monitor.")



"""Module for confirmation for setting UTC time."""
from pyvlx.const import Command

from .frame import FrameBase


class FrameSetUTCConfirmation(FrameBase):
    """Frame for confirmation for setting UTC time."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_SET_UTC_CFM)



"""Module for retrieving scene list from API."""
from pyvlx.api_event import ApiEvent
from pyvlx.frames import (CommandSendConfirmationStatus,
                          FrameCommandRemainingTimeNotification,
                          FrameCommandRunStatusNotification,
                          FrameCommandSendConfirmation,
                          FrameCommandSendRequest,
                          FrameSessionFinishedNotification)
from pyvlx.session_id import get_new_session_id


class CommandSend(ApiEvent):
    """Class for sending command to API."""

    def __init__(self, pyvlx, node_id, position, wait_for_completion=True, timeout_in_seconds=60):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx, timeout_in_seconds=timeout_in_seconds)
        self.success = False
        self.node_id = node_id
        self.position = position
        self.wait_for_completion = wait_for_completion
        self.session_id = None

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameCommandSendConfirmation) and frame.session_id == self.session_id:
            if frame.status == CommandSendConfirmationStatus.ACCEPTED:
                self.success = True
            return not self.wait_for_completion
        if isinstance(frame, FrameCommandRemainingTimeNotification) and frame.session_id == self.session_id:
            # Ignoring FrameCommandRemainingTimeNotification
            return False
        if isinstance(frame, FrameCommandRunStatusNotification) and frame.session_id == self.session_id:
            # At the moment I don't reall understand what the FrameCommandRunStatusNotification is good for.
            # Ignoring these packets for now
            return False
        if isinstance(frame, FrameSessionFinishedNotification) and frame.session_id == self.session_id:
            return True
        return False

    def request_frame(self):
        """Construct initiating frame."""
        self.session_id = get_new_session_id()
        return FrameCommandSendRequest(node_ids=[self.node_id], position=self.position, session_id=self.session_id)



"""Helper module for SLIP Frames."""
from pyvlx.const import Command
from pyvlx.exception import PyVLXException


def calc_crc(raw):
    """Calculate cyclic redundancy check (CRC)."""
    crc = 0
    for sym in raw:
        crc = crc ^ int(sym)
    return crc


def extract_from_frame(data):
    """Extract payload and command from frame."""
    if len(data) <= 4:
        raise PyVLXException("could_not_extract_from_frame_too_short", data=data)
    length = data[0] * 256 + data[1] - 1
    if len(data) != length + 3:
        raise PyVLXException("could_not_extract_from_frame_invalid_length", data=data, current_length=len(data), expected_length=length + 3)
    if calc_crc(data[:-1]) != data[-1]:
        raise PyVLXException("could_not_extract_from_frame_invalid_crc", data=data, expected_crc=calc_crc(data[:-1]), current_crc=data[-1])
    payload = data[4:-1]
    try:
        command = Command(data[2] * 256 + data[3])
    except ValueError:
        raise PyVLXException("could_not_extract_from_frame_command", data=data)
    return command, payload



"""Module for storing alias array."""
from .exception import PyVLXException


class AliasArray():
    """Object for storing alias array."""

    def __init__(self, raw=None):
        """Initialize alias array."""
        self.alias_array_ = []
        if raw is not None:
            self.parse_raw(raw)

    def __str__(self):
        """Return human readable string."""
        return ", ".join("{:02x}{:02x}={:02x}{:02x}".format(c[0][0], c[0][1], c[1][0], c[1][1]) for c in self.alias_array_)

    def __bytes__(self):
        """Get raw bytes of alias array."""
        ret = bytes([len(self.alias_array_)])
        for alias in self.alias_array_:
            ret += alias[0] + alias[1]
        ret += bytes((5-len(self.alias_array_))*4)
        return ret

    def parse_raw(self, raw):
        """Parse alias array from raw bytes."""
        if not isinstance(raw, bytes):
            raise PyVLXException("AliasArray::invalid_type_if_raw", type_raw=type(raw))
        if len(raw) != 21:
            raise PyVLXException("AliasArray::invalid_size", size=len(raw))
        nbr_of_alias = raw[0]
        if nbr_of_alias > 5:
            raise PyVLXException("AliasArray::invalid_nbr_of_alias", nbr_of_alias=nbr_of_alias)
        for i in range(0, nbr_of_alias):
            self.alias_array_.append((raw[i*4+1:i*4+3], raw[i*4+3:i*4+5]))



"""Module for error notification."""
from enum import Enum

from pyvlx.const import Command

from .frame import FrameBase


class ErrorType(Enum):
    """Enum class for error types."""

    NotFurtherDefined = 0
    UnknownCommand = 1
    ErrorOnFrameStructure = 2
    BusBusy = 7
    BadSystemTableIndex = 8
    NotAuthenticated = 12


class FrameErrorNotification(FrameBase):
    """Frame for error notification."""

    PAYLOAD_LEN = 1

    def __init__(self, error_type=ErrorType.NotFurtherDefined):
        """Init Frame."""
        super().__init__(Command.GW_ERROR_NTF)
        self.error_type = error_type

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.error_type.value])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.error_type = ErrorType(payload[0])

    def __str__(self):
        """Return human readable string."""
        return '<FrameErrorNotification error_type=\'{}\'/>'.format(self.error_type)



"""Helper module for creating a frame out of raw data."""
from pyvlx.frames import (FrameActivateSceneConfirmation,
                          FrameActivateSceneRequest,
                          FrameActivationLogUpdatedNotification,
                          FrameCommandRemainingTimeNotification,
                          FrameCommandRunStatusNotification,
                          FrameCommandSendConfirmation,
                          FrameCommandSendRequest,
                          FrameDiscoverNodesConfirmation,
                          FrameDiscoverNodesNotification,
                          FrameDiscoverNodesRequest, FrameErrorNotification,
                          FrameGetAllNodesInformationConfirmation,
                          FrameGetAllNodesInformationFinishedNotification,
                          FrameGetAllNodesInformationNotification,
                          FrameGetAllNodesInformationRequest,
                          FrameGetNodeInformationConfirmation,
                          FrameGetNodeInformationNotification,
                          FrameGetNodeInformationRequest,
                          FrameGetProtocolVersionConfirmation,
                          FrameGetProtocolVersionRequest,
                          FrameGetSceneListConfirmation,
                          FrameGetSceneListNotification,
                          FrameGetSceneListRequest, FrameGetStateConfirmation,
                          FrameGetStateRequest, FrameGetVersionConfirmation,
                          FrameGetVersionRequest,
                          FrameHouseStatusMonitorDisableConfirmation,
                          FrameHouseStatusMonitorDisableRequest,
                          FrameHouseStatusMonitorEnableConfirmation,
                          FrameHouseStatusMonitorEnableRequest,
                          FrameNodeInformationChangedNotification,
                          FrameNodeStatePositionChangedNotification,
                          FramePasswordEnterConfirmation,
                          FramePasswordEnterRequest,
                          FrameSessionFinishedNotification,
                          FrameSetNodeNameConfirmation,
                          FrameSetNodeNameRequest, FrameSetUTCConfirmation,
                          FrameSetUTCRequest, extract_from_frame)

from .const import Command
from .log import PYVLXLOG


def frame_from_raw(raw):
    """Create and return frame from raw bytes."""
    command, payload = extract_from_frame(raw)
    frame = create_frame(command)
    if frame is None:
        PYVLXLOG.warning("Command %s not implemented, raw: %s", command, ":".join("{:02x}".format(c) for c in raw))
        return None
    frame.validate_payload_len(payload)
    frame.from_payload(payload)
    return frame


def create_frame(command):
    """Create and return empty Frame from Command."""
    # pylint: disable=too-many-branches,too-many-return-statements
    if command == Command.GW_ERROR_NTF:
        return FrameErrorNotification()
    if command == Command.GW_COMMAND_SEND_REQ:
        return FrameCommandSendRequest()
    if command == Command.GW_COMMAND_SEND_CFM:
        return FrameCommandSendConfirmation()
    if command == Command.GW_COMMAND_RUN_STATUS_NTF:
        return FrameCommandRunStatusNotification()
    if command == Command.GW_COMMAND_REMAINING_TIME_NTF:
        return FrameCommandRemainingTimeNotification()
    if command == Command.GW_SESSION_FINISHED_NTF:
        return FrameSessionFinishedNotification()

    if command == Command.GW_PASSWORD_ENTER_REQ:
        return FramePasswordEnterRequest()
    if command == Command.GW_PASSWORD_ENTER_CFM:
        return FramePasswordEnterConfirmation()

    if command == Command.GW_CS_DISCOVER_NODES_REQ:
        return FrameDiscoverNodesRequest()
    if command == Command.GW_CS_DISCOVER_NODES_CFM:
        return FrameDiscoverNodesConfirmation()
    if command == Command.GW_CS_DISCOVER_NODES_NTF:
        return FrameDiscoverNodesNotification()

    if command == Command.GW_GET_SCENE_LIST_REQ:
        return FrameGetSceneListRequest()
    if command == Command.GW_GET_SCENE_LIST_CFM:
        return FrameGetSceneListConfirmation()
    if command == Command.GW_GET_SCENE_LIST_NTF:
        return FrameGetSceneListNotification()

    if command == Command.GW_GET_NODE_INFORMATION_REQ:
        return FrameGetNodeInformationRequest()
    if command == Command.GW_GET_NODE_INFORMATION_CFM:
        return FrameGetNodeInformationConfirmation()
    if command == Command.GW_GET_NODE_INFORMATION_NTF:
        return FrameGetNodeInformationNotification()

    if command == Command.GW_GET_ALL_NODES_INFORMATION_REQ:
        return FrameGetAllNodesInformationRequest()
    if command == Command.GW_GET_ALL_NODES_INFORMATION_CFM:
        return FrameGetAllNodesInformationConfirmation()
    if command == Command.GW_GET_ALL_NODES_INFORMATION_NTF:
        return FrameGetAllNodesInformationNotification()
    if command == Command.GW_GET_ALL_NODES_INFORMATION_FINISHED_NTF:
        return FrameGetAllNodesInformationFinishedNotification()

    if command == Command.GW_ACTIVATE_SCENE_REQ:
        return FrameActivateSceneRequest()
    if command == Command.GW_ACTIVATE_SCENE_CFM:
        return FrameActivateSceneConfirmation()

    if command == Command.GW_GET_VERSION_REQ:
        return FrameGetVersionRequest()
    if command == Command.GW_GET_VERSION_CFM:
        return FrameGetVersionConfirmation()
    if command == Command.GW_GET_PROTOCOL_VERSION_REQ:
        return FrameGetProtocolVersionRequest()
    if command == Command.GW_GET_PROTOCOL_VERSION_CFM:
        return FrameGetProtocolVersionConfirmation()

    if command == Command.GW_SET_NODE_NAME_REQ:
        return FrameSetNodeNameRequest()
    if command == Command.GW_SET_NODE_NAME_CFM:
        return FrameSetNodeNameConfirmation()

    if command == Command.GW_NODE_INFORMATION_CHANGED_NTF:
        return FrameNodeInformationChangedNotification()

    if command == Command.GW_GET_STATE_REQ:
        return FrameGetStateRequest()
    if command == Command.GW_GET_STATE_CFM:
        return FrameGetStateConfirmation()

    if command == Command.GW_SET_UTC_REQ:
        return FrameSetUTCRequest()
    if command == Command.GW_SET_UTC_CFM:
        return FrameSetUTCConfirmation()

    if command == Command.GW_ACTIVATION_LOG_UPDATED_NTF:
        return FrameActivationLogUpdatedNotification()

    if command == Command.GW_HOUSE_STATUS_MONITOR_ENABLE_REQ:
        return FrameHouseStatusMonitorEnableRequest()
    if command == Command.GW_HOUSE_STATUS_MONITOR_ENABLE_CFM:
        return FrameHouseStatusMonitorEnableConfirmation()
    if command == Command.GW_HOUSE_STATUS_MONITOR_DISABLE_REQ:
        return FrameHouseStatusMonitorDisableRequest()
    if command == Command.GW_HOUSE_STATUS_MONITOR_DISABLE_CFM:
        return FrameHouseStatusMonitorDisableConfirmation()

    if command == Command.GW_NODE_STATE_POSITION_CHANGED_NTF:
        return FrameNodeStatePositionChangedNotification()

    return None



"""
Module for basic object for nodes.

Node object is an interface class and should
be derived by other objects like window openers
and roller shutters.
"""
from .exception import PyVLXException
from .set_node_name import SetNodeName


class Node:
    """Class for node abstraction."""

    def __init__(self, pyvlx, node_id, name):
        """Initialize Node object."""
        self.pyvlx = pyvlx
        self.node_id = node_id
        self.name = name
        self.device_updated_cbs = []

    def register_device_updated_cb(self, device_updated_cb):
        """Register device updated callback."""
        self.device_updated_cbs.append(device_updated_cb)

    def unregister_device_updated_cb(self, device_updated_cb):
        """Unregister device updated callback."""
        self.device_updated_cbs.remove(device_updated_cb)

    async def after_update(self):
        """Execute callbacks after internal state has been changed."""
        for device_updated_cb in self.device_updated_cbs:
            # pylint: disable=not-callable
            await device_updated_cb(self)

    async def rename(self, name):
        """Change name of node."""
        set_node_name = SetNodeName(pyvlx=self.pyvlx, node_id=self.node_id, name=name)
        await set_node_name.do_api_call()
        if not set_node_name.success:
            raise PyVLXException("Unable to rename node")
        self.name = name

    def __str__(self):
        """Return object as readable string."""
        return '<{} name="{}" ' \
            'node_id="{}"/>' \
            .format(
                type(self).__name__,
                self.name,
                self.node_id)

    def __eq__(self, other):
        """Equal operator."""
        return type(self).__name__ == type(other).__name__ and \
            self.__dict__ == other.__dict__



"""Module for string encoding, decoding."""
from .exception import PyVLXException


def string_to_bytes(string, size):
    """Convert string to bytes add padding."""
    if len(string) > size:
        raise PyVLXException("string_to_bytes::string_to_large")
    encoded = bytes(string, encoding='utf-8')
    return encoded + bytes(size-len(encoded))


def bytes_to_string(raw):
    """Convert bytes to string."""
    ret = bytes()
    for byte in raw:
        if byte == 0x00:
            return ret.decode("utf-8")
        ret += bytes([byte])
    return ret.decode("utf-8")



"""Module for sending command to gw."""
from enum import Enum

from pyvlx.const import Command, Originator, Priority
from pyvlx.exception import PyVLXException
from pyvlx.position import Position

from .frame import FrameBase


class FrameCommandSendRequest(FrameBase):
    """Frame for sending command to gw."""

    PAYLOAD_LEN = 66

    def __init__(self, node_ids=None, position=Position(), session_id=None, originator=Originator.USER):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_SEND_REQ)
        self.node_ids = node_ids
        self.position = position
        self.session_id = session_id
        self.originator = originator
        self.priority = Priority.USER_LEVEL_2

    def get_payload(self):
        """Return Payload."""
        # Session id
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.originator.value])
        ret += bytes([self.priority.value])
        ret += bytes([0])  # ParameterActive pointing to main parameter (MP)
        # FPI 1+2
        ret += bytes([0])
        ret += bytes([0])

        # Main parameter + functional parameter (in our case: position)
        ret += bytes(self.position)
        ret += bytes(32)

        # Nodes array: Number of nodes + node array + padding
        ret += bytes([len(self.node_ids)])  # index array count
        ret += bytes(self.node_ids) + bytes(20-len(self.node_ids))

        # Priority  Level Lock
        ret += bytes([0])
        # Priority Level information 1+2
        ret += bytes([0, 0])
        # Locktime
        ret += bytes([0])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0]*256 + payload[1]
        self.originator = Originator(payload[2])
        self.priority = Priority(payload[3])

        len_node_ids = payload[41]
        if len_node_ids > 20:
            raise PyVLXException("command_send_request_wrong_node_length")
        self.node_ids = []
        for i in range(len_node_ids):
            self.node_ids.append(payload[42] + i)

        self.position = int(payload[7]/2)
        if self.position > 100:
            raise PyVLXException("command_send_request_wrong_position")

    def __str__(self):
        """Return human readable string."""
        return '<FrameCommandSendRequest node_ids={} position="{}" session_id={} originator={}/>'.format(
            self.node_ids, self.position, self.session_id,
            self.originator)


class CommandSendConfirmationStatus(Enum):
    """Enum class for status of command send confirmation."""

    REJECTED = 0
    ACCEPTED = 1


class FrameCommandSendConfirmation(FrameBase):
    """Frame for confirmation of command send frame."""

    PAYLOAD_LEN = 3

    def __init__(self, session_id=None, status=None):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_SEND_CFM)
        self.session_id = session_id
        self.status = status

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.status.value])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0]*256 + payload[1]
        self.status = CommandSendConfirmationStatus(payload[2])

    def __str__(self):
        """Return human readable string."""
        return '<FrameCommandSendConfirmation session_id={} status={}/>'.format(self.session_id, self.status)


class FrameCommandRunStatusNotification(FrameBase):
    """Frame for run status notification in scope of command send frame."""

    PAYLOAD_LEN = 13

    def __init__(self, session_id=None, status_id=None, index_id=None, node_parameter=None, parameter_value=None):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_RUN_STATUS_NTF)
        self.session_id = session_id
        self.status_id = status_id
        self.index_id = index_id
        self.node_parameter = node_parameter
        self.parameter_value = parameter_value

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.status_id])
        ret += bytes([self.index_id])
        ret += bytes([self.node_parameter])
        ret += bytes([self.parameter_value >> 8 & 255, self.parameter_value & 255])

        # XXX: Missing implementation of run_status, status_reply and information_code
        ret += bytes(6)
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0]*256 + payload[1]
        self.status_id = payload[2]
        self.index_id = payload[3]
        self.node_parameter = payload[4]
        self.parameter_value = payload[5]*256 + payload[6]

    def __str__(self):
        """Return human readable string."""
        return \
            '<FrameCommandRunStatusNotification session_id={} status_id={} ' \
            'index_id={} node_parameter={} parameter_value={}/>'.format(
                self.session_id, self.status_id, self.index_id,
                self.node_parameter, self.parameter_value)


class FrameCommandRemainingTimeNotification(FrameBase):
    """Frame for notification of remaining time in scope of command send frame."""

    PAYLOAD_LEN = 6

    def __init__(self, session_id=None, index_id=None, node_parameter=None, seconds=0):
        """Init Frame."""
        super().__init__(Command.GW_COMMAND_REMAINING_TIME_NTF)
        self.session_id = session_id
        self.index_id = index_id
        self.node_parameter = node_parameter
        self.seconds = seconds

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.index_id])
        ret += bytes([self.node_parameter])
        ret += bytes([self.seconds >> 8 & 255, self.seconds & 255])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0]*256 + payload[1]
        self.index_id = payload[2]
        self.node_parameter = payload[3]
        self.seconds = payload[4]*256 + payload[5]

    def __str__(self):
        """Return human readable string."""
        return \
            '<FrameCommandRemainingTimeNotification session_id={} index_id={} ' \
            'node_parameter={} seconds={}/>'.format(
                self.session_id, self.index_id, self.node_parameter, self.seconds)


class FrameSessionFinishedNotification(FrameBase):
    """Frame for notification of session finishid in scope of command send frame."""

    PAYLOAD_LEN = 2

    def __init__(self, session_id=None):
        """Init Frame."""
        super().__init__(Command.GW_SESSION_FINISHED_NTF)
        self.session_id = session_id

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0]*256 + payload[1]

    def __str__(self):
        """Return human readable string."""
        return '<FrameSessionFinishedNotification session_id={} />'.format(self.session_id)



"""Module for error notification."""
from pyvlx.const import Command

from .frame import FrameBase


class FrameActivationLogUpdatedNotification(FrameBase):
    """Frame for error notification."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_ACTIVATION_LOG_UPDATED_NTF)



"""Module for confirmation for  disabling the house status monitor."""
from pyvlx.const import Command

from .frame import FrameBase


class FrameHouseStatusMonitorDisableConfirmation(FrameBase):
    """Frame for requesting enabling the house status monitor."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_HOUSE_STATUS_MONITOR_DISABLE_CFM)



"""Module for sending get state requests to API in regular periods."""
import asyncio

from .exception import PyVLXException
from .get_state import GetState


class Heartbeat():
    """Class for sending heartbeats to API."""

    def __init__(self, pyvlx, timeout_in_seconds=60):
        """Initialize Heartbeat object."""
        self.pyvlx = pyvlx
        self.timeout_in_seconds = timeout_in_seconds
        self.loop_event = asyncio.Event()
        self.stopped = False
        self.run_task = None
        self.timeout_handle = None
        self.stopped_event = asyncio.Event()

    def __del__(self):
        """Cleanup heartbeat."""
        self.cancel_loop_timeout()

    def start(self):
        """Create loop task."""
        self.run_task = self.pyvlx.loop.create_task(
            self.loop())

    async def stop(self):
        """Stop heartbeat."""
        self.stopped = True
        self.loop_event.set()
        # Waiting for shutdown of loop()
        await self.stopped_event.wait()

    async def loop(self):
        """Pulse every timeout seconds until stopped."""
        while not self.stopped:
            self.timeout_handle = self.pyvlx.connection.loop.call_later(
                self.timeout_in_seconds, self.loop_timeout)
            await self.loop_event.wait()
            if not self.stopped:
                self.loop_event.clear()
                await self.pulse()
        self.cancel_loop_timeout()
        self.stopped_event.set()

    def loop_timeout(self):
        """Handle loop timeout."""
        self.loop_event.set()

    def cancel_loop_timeout(self):
        """Cancel loop timeout."""
        if self.timeout_handle is not None:
            self.timeout_handle.cancel()
            self.timeout_handle = None

    async def pulse(self):
        """Send get state request to API to keep the connection alive."""
        get_state = GetState(pyvlx=self.pyvlx)
        await get_state.do_api_call()
        if not get_state.success:
            raise PyVLXException("Unable to send get state.")



"""Frames for receiving state from gateway."""
from enum import Enum

from pyvlx.const import Command

from .frame import FrameBase


class FrameGetStateRequest(FrameBase):
    """Frame for requesting state."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_STATE_REQ)


class GatewayState(Enum):
    """Enum class for status if gateway."""

    TEST_MODE = 0
    GATEWAY_MODE_NO_ACTUATOR = 1
    GATEWAY_MODE_WITH_ACTUATORS = 2
    BEACON_MODE_NOT_CONFIGURED = 3
    BEACON_MODE_CONFIGURED = 4


class GatewaySubState(Enum):
    """Enum class for substate if gateway."""

    IDLE = 0x00
    PERFORMING_TASK_CONFIGURATION_SERVICE_HANDLER = 0x01
    PERFORMING_TASK_SCENE_CONFIGURATION = 0x02
    PERFORMING_TASK_INFORMATION_SERVICE_CONFIGURATION = 0x03
    PERFORMING_TASK_CONTACT_INPUT_CONFIGURATION = 0x04
    PERFORMING_TASK_COMMAND = 0x80
    PERFORMING_TASK_ACTIVATE_GROUP = 0x81
    PERFORMING_TASK_ACTIVATE_SCENE = 0x82


class FrameGetStateConfirmation(FrameBase):
    """Frame for confirmation for get state requests."""

    PAYLOAD_LEN = 6

    def __init__(self, gateway_state=GatewayState.TEST_MODE, gateway_sub_state=GatewaySubState.IDLE):
        """Init Frame."""
        super().__init__(Command.GW_GET_STATE_CFM)
        self.gateway_state = gateway_state
        self.gateway_sub_state = gateway_sub_state

    def get_payload(self):
        """Return Payload."""
        payload = bytes([self.gateway_state.value, self.gateway_sub_state.value])
        payload += bytes(4)  # State date, reserved for future use
        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.gateway_state = GatewayState(payload[0])
        self.gateway_sub_state = GatewaySubState(payload[1])

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetStateConfirmation gateway_state="{}" gateway_sub_state="{}"/>'.format(self.gateway_state, self.gateway_sub_state)



"""
Module for PyVLX object.

PyVLX is an asynchronous library for connecting to
a VELUX KLF 200 device for controlling window openers
and roller shutters.
"""
import asyncio

from .config import Config
from .connection import Connection
from .exception import PyVLXException
from .get_protocol_version import GetProtocolVersion
from .get_version import GetVersion
from .heartbeat import Heartbeat
from .house_status_monitor import house_status_monitor_enable
from .log import PYVLXLOG
from .login import Login
from .nodes import Nodes
from .scenes import Scenes
from .set_utc import set_utc
from .node_updater import update_nodes


class PyVLX:
    """Class for PyVLX."""

    def __init__(self, path=None, host=None, password=None, log_frames=False, loop=None):
        """Initialize PyVLX class."""
        self.loop = loop or asyncio.get_event_loop()
        self.config = Config(self, path, host, password)
        self.connection = Connection(loop=self.loop, config=self.config)
        self.heartbeat = Heartbeat(pyvlx=self)
        self.heartbeat.start()
        if log_frames:
            self.connection.register_frame_received_cb(self.log_frame)
        self.connection.register_frame_received_cb(update_nodes)
        self.nodes = Nodes(self)
        self.scenes = Scenes(self)
        self.version = None
        self.protocol_version = None

    async def connect(self):
        """Connect to KLF 200."""
        PYVLXLOG.warning("Connecting to KLF 200.")
        await self.connection.connect()
        login = Login(pyvlx=self, password=self.config.password)
        await login.do_api_call()
        if not login.success:
            raise PyVLXException("Unable to login")

    async def update_version(self):
        """Retrieve version and protocol version from API."""
        get_version = GetVersion(pyvlx=self)
        await get_version.do_api_call()
        if not get_version.success:
            raise PyVLXException("Unable to retrieve version")
        self.version = get_version.version
        get_protocol_version = GetProtocolVersion(pyvlx=self)
        await get_protocol_version.do_api_call()
        if not get_protocol_version.success:
            raise PyVLXException("Unable to retrieve protocol version")
        self.protocol_version = get_protocol_version.version
        PYVLXLOG.warning(
            "Connected to: %s, protocol version: %s",
            self.version, self.protocol_version)

    async def send_frame(self, frame):
        """Send frame to API via connection."""
        if not self.connection.connected:
            await self.connect()
            await self.update_version()
            await set_utc(pyvlx=self)
            await house_status_monitor_enable(pyvlx=self)
        self.connection.write(frame)

    async def disconnect(self):
        """Disconnect from KLF 200."""
        await self.heartbeat.stop()
        self.connection.disconnect()

    async def load_nodes(self, node_id=None):
        """Load devices from KLF 200, if no node_id is specified all nodes are loaded."""
        await self.nodes.load(node_id)

    async def load_scenes(self):
        """Load scenes from KLF 200."""
        await self.scenes.load()

    async def log_frame(self, frame):
        """Log frame to logger."""
        PYVLXLOG.info("%s", frame)



"""Module for handling the TCP connection with Gateway."""
import asyncio
import ssl

from .exception import PyVLXException
from .frame_creation import frame_from_raw
from .frames import FrameBase
from .slip import get_next_slip, is_slip, slip_pack


class SlipTokenizer:
    """Helper class for splitting up binary stream to slip packets."""

    def __init__(self):
        """Init Tokenizer."""
        self.data = bytes()

    def feed(self, chunk):
        """Feed chunk to tokenizer."""
        if not chunk:
            return
        self.data += chunk

    def has_tokens(self):
        """Return True if Tokenizer has tokens."""
        return is_slip(self.data)

    def get_next_token(self):
        """Get next token from Tokenizer."""
        slip, self.data = get_next_slip(self.data)
        return slip


class TCPTransport(asyncio.Protocol):
    """Class for handling asyncio connection transport."""

    def __init__(self, frame_received_cb, connection_closed_cb):
        """Init TCPTransport."""
        self.frame_received_cb = frame_received_cb
        self.connection_closed_cb = connection_closed_cb
        self.tokenizer = SlipTokenizer()

    def connection_made(self, transport):
        """Handle sucessful connection."""

    def data_received(self, data):
        """Handle data received."""
        self.tokenizer.feed(data)
        while self.tokenizer.has_tokens():
            raw = self.tokenizer.get_next_token()
            frame = frame_from_raw(raw)
            if frame is not None:
                self.frame_received_cb(frame)

    def connection_lost(self, exc):
        """Handle lost connection."""
        self.connection_closed_cb()


class Connection:
    """Class for handling TCP connection."""

    def __init__(self, loop, config):
        """Init TCP connection."""
        self.loop = loop
        self.config = config
        self.transport = None
        self.frame_received_cbs = []
        self.connected = False

    def __del__(self):
        """Destruct connection."""
        self.disconnect()

    def disconnect(self):
        """Disconnect connection."""
        if self.transport is not None:
            self.transport.close()
            self.transport = None

    async def connect(self):
        """Connect to gateway via SSL."""
        tcp_client = TCPTransport(self.frame_received_cb, self.connection_closed_cb)
        self.transport, _ = await self.loop.create_connection(
            lambda: tcp_client,
            host=self.config.host,
            port=self.config.port,
            ssl=self.create_ssl_context())
        self.connected = True

    def register_frame_received_cb(self, callback):
        """Register frame received callback."""
        self.frame_received_cbs.append(callback)

    def unregister_frame_received_cb(self, callback):
        """Unregister frame received callback."""
        self.frame_received_cbs.remove(callback)

    def write(self, frame):
        """Write frame to Bus."""
        if not isinstance(frame, FrameBase):
            raise PyVLXException("Frame not of type FrameBase", frame_type=type(frame))
        self.transport.write(slip_pack(bytes(frame)))

    @staticmethod
    def create_ssl_context():
        """Create and return SSL Context."""
        ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
        return ssl_context

    def frame_received_cb(self, frame):
        """Received message."""
        for frame_received_cb in self.frame_received_cbs:
            # pylint: disable=not-callable
            self.loop.create_task(frame_received_cb(frame))

    def connection_closed_cb(self):
        """Server closed connection."""
        self.connected = False



"""Module for generating a unique session_id."""


LAST_SESSION_ID = 0
MAX_SESSION_ID = 65535


def get_new_session_id():
    """Generate new session_id."""
    global LAST_SESSION_ID  # pylint: disable=global-statement
    LAST_SESSION_ID = LAST_SESSION_ID + 1
    if LAST_SESSION_ID > MAX_SESSION_ID:
        LAST_SESSION_ID = 1
    return LAST_SESSION_ID


def set_session_id(session_id):
    """Set session id to value."""
    global LAST_SESSION_ID  # pylint: disable=global-statement
    LAST_SESSION_ID = session_id



"""Module for retrieving gateway state from API."""
from .api_event import ApiEvent
from .frames import FrameGetStateConfirmation, FrameGetStateRequest


class GetState(ApiEvent):
    """Class for retrieving gateway state from API."""

    def __init__(self, pyvlx):
        """Initialize login class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False
        self.gateway_state = None
        self.gateway_sub_state = None

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameGetStateConfirmation):
            return False
        self.success = True
        self.gateway_state = frame.gateway_state
        self.gateway_sub_state = frame.gateway_sub_state
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetStateRequest()



"""Module for enum and consts."""

from enum import Enum


class Command(Enum):
    """Enum class for GW Command bytes."""

    # pylint: disable=invalid-name

    GW_ERROR_NTF = 0x0000
    GW_REBOOT_REQ = 0x0001
    GW_REBOOT_CFM = 0x0002

    GW_SET_FACTORY_DEFAULT_REQ = 0x0003
    GW_SET_FACTORY_DEFAULT_CFM = 0x0004

    GW_GET_VERSION_REQ = 0x0008
    GW_GET_VERSION_CFM = 0x0009

    GW_GET_PROTOCOL_VERSION_REQ = 0x000A
    GW_GET_PROTOCOL_VERSION_CFM = 0x000B

    GW_GET_STATE_REQ = 0x000C
    GW_GET_STATE_CFM = 0x000D

    GW_LEAVE_LEARN_STATE_REQ = 0x000E
    GW_LEAVE_LEARN_STATE_CFM = 0x000F

    GW_GET_NETWORK_SETUP_REQ = 0x00E0
    GE_GET_NETWORK_SETUP_CFM = 0x00E1
    GW_SET_NETWORK_SETUP_REQ = 0x00E2
    GE_SET_NETWORK_SETUP_CFM = 0x00E3

    GW_CS_GET_SYSTEMTABLE_DATQ_REQ = 0x0100
    GW_CS_GET_SYSTEMTABLE_DATA_CFM = 0x0101
    GW_CS_GET_SYSTEMTABLE_DATA_NTF = 0x0102

    GW_CS_DISCOVER_NODES_REQ = 0x0103
    GW_CS_DISCOVER_NODES_CFM = 0x0104
    GW_CS_DISCOVER_NODES_NTF = 0x0105

    GW_CS_REMOVE_NODES_REQ = 0x0106
    GW_CS_REMOVE_NODES_CFM = 0x0107

    GW_CS_VIRGIN_STATE_REQ = 0x0108
    GW_CS_VIRGIN_STATE_CFM = 0x0109

    GW_CS_CONTROLLER_COPY_REQ = 0x010A
    GW_CS_CONTROLLER_COPY_CFM = 0x010B
    GW_CS_CONTROLLER_COPY_NTF = 0x010C
    GW_CS_CONTROLLER_COPY_CANCEL_NTF = 0x010D

    GW_CS_RECEIVE_KEY_REQ = 0x010E
    GW_CS_RECEIVE_KEY_CFM = 0x010F
    GW_CS_RECEIVE_KEY_NTF = 0x0110

    GW_CS_PGC_JOB_NTF = 0x0111
    GW_CS_SYSTEM_TABLE_UPDATE_NTF = 0x0112
    GW_CS_GENERATE_NEW_KEY_REQ = 0x0113
    GW_CS_GENERATE_NEW_KEY_CFM = 0x0114
    GW_CS_GENERATE_NEW_KEY_NTF = 0x0115

    GW_CS_REPAIR_KEY_REQ = 0x0116
    GW_CS_REPAIR_KEY_CFM = 0x0117
    GW_CS_REPAID_KEY_NTF = 0x0118

    GW_CS_ACTIVATE_CONFIGURATION_MODE_REQ = 0x0119
    GW_CS_ACTIVATE_CONFIGURATION_MODE_CFM = 0x011A

    GW_GET_NODE_INFORMATION_REQ = 0x0200
    GW_GET_NODE_INFORMATION_CFM = 0x0201
    GW_GET_NODE_INFORMATION_NTF = 0x0210

    GW_GET_ALL_NODES_INFORMATION_REQ = 0x0202
    GW_GET_ALL_NODES_INFORMATION_CFM = 0x0203
    GW_GET_ALL_NODES_INFORMATION_NTF = 0x0204
    GW_GET_ALL_NODES_INFORMATION_FINISHED_NTF = 0x0205

    GW_SET_NODE_VARIATION_REQ = 0x0206
    GW_SET_NODE_VARIATION_CFM = 0x0207

    GW_SET_NODE_NAME_REQ = 0x0208
    GW_SET_NODE_NAME_CFM = 0x0209

    GW_SET_NODE_VELOCITY_REQ = 0x020A
    GW_SET_NODE_VELOCITY_CFM = 0x020B

    GW_NODE_INFORMATION_CHANGED_NTF = 0x020C

    GW_NODE_STATE_POSITION_CHANGED_NTF = 0x0211

    GW_SET_NODE_ORDER_AND_PLACEMENT_REQ = 0x020D
    GW_SET_NODE_ORDER_AND_PLACEMENT_CFM = 0x020E

    GW_GET_GROUP_INFORMATION_REQ = 0x0220
    GW_GET_GROUP_INFORMATION_CFM = 0x0221
    GW_GET_GROUP_INFORMATION_NTF = 0x0230

    GW_SET_GROUP_INFORMATION_REQ = 0x0222
    GW_SET_GROUP_INFORMATION_CFM = 0x0223

    GW_GROUP_INFORMATION_CHANGED_NTF = 0x0224

    GW_DELETE_GROUP_REQ = 0x0225
    GW_DELETE_GROUP_CFM = 0x0226

    GW_NEW_GROUP_REQ = 0x0227
    GW_NEW_GROUP_CFM = 0x0228

    GW_GET_ALL_GROUPS_INFORMATION_REQ = 0x0229
    GW_GET_ALL_GROUPS_INFORMATION_CFM = 0x022A
    GW_GET_ALL_GROUPS_INFORMATION_NTF = 0x022B

    GW_GET_ALL_GROUPS_INFORMATION_FINISHED_NTF = 0x022C

    GW_GROUP_DELETED_NTF = 0x022D

    GW_HOUSE_STATUS_MONITOR_ENABLE_REQ = 0x0240
    GW_HOUSE_STATUS_MONITOR_ENABLE_CFM = 0x0241

    GW_HOUSE_STATUS_MONITOR_DISABLE_REQ = 0x0242
    GW_HOUSE_STATUS_MONITOR_DISABLE_CFM = 0x0243

    GW_COMMAND_SEND_REQ = 0x0300
    GW_COMMAND_SEND_CFM = 0x0301
    GW_COMMAND_RUN_STATUS_NTF = 0x0302
    GW_COMMAND_REMAINING_TIME_NTF = 0x0303
    GW_SESSION_FINISHED_NTF = 0x0304

    GW_STATUS_REQUEST_REQ = 0x0305
    GW_STATUS_REQUEST_CFM = 0x0306
    GW_STATUS_REQUEST_NTF = 0x0307

    GW_WINK_SEND_REQ = 0x0308
    GW_WINK_SEND_CFM = 0x0309
    GW_WINK_SEND_NTF = 0x030A

    GW_SET_LIMITATION_REQ = 0x0310
    GW_SET_LIMITATION_CFM = 0x0311
    GW_GET_LIMITATION_STATUS_REQ = 0x0312
    GW_GET_LIMITATION_STATUS_CFM = 0x0313
    GW_LIMITATION_STATUS_NTF = 0x0314

    GW_MODE_SEND_REQ = 0x0320
    GW_MODE_SEND_CFM = 0x0321
    GW_MODE_SEND_NTF = 0x0322

    GW_INITIALIZE_SCENE_REQ = 0x0400
    GW_INITIALIZE_SCENE_CFM = 0x0401
    GW_INITIALIZE_SCENE_NTF = 0x0402
    GW_INITIALIZE_SCENE_CANCEL_REQ = 0x0403
    GW_INITIALIZE_SCENE_CANCEL_CFM = 0x0404
    GW_RECORD_SCENE_REQ = 0x0405
    GW_RECORD_SCENE_CFM = 0x0406
    GW_RECORD_SCENE_NTF = 0x0407

    GW_DELETE_SCENE_REQ = 0x0408
    GW_DELETE_SCENE_CFM = 0x0409

    GW_RENAME_SCENE_REQ = 0x040A
    GW_RENAME_SCENE_CFM = 0x040B

    GW_GET_SCENE_LIST_REQ = 0x040C
    GW_GET_SCENE_LIST_CFM = 0x040D
    GW_GET_SCENE_LIST_NTF = 0x040E

    GW_GET_SCENE_INFORMATION_REQ = 0x040F
    GW_GET_SCENE_INFORMATION_CFM = 0x0410
    GW_GET_SCENE_INFORMATION_NTF = 0x0411

    GW_ACTIVATE_SCENE_REQ = 0x0412
    GW_ACTIVATE_SCENE_CFM = 0x0413

    GW_STOP_SCENE_REQ = 0x0415
    GW_STOP_SCENE_CFM = 0x0416

    GW_SCENE_INFORMATION_CHANGED_NTF = 0x0419

    GW_ACTIVATE_PRODUCTGROUP_REQ = 0x0447
    GW_ACTIVATE_PRODUCTGROUP_CFM = 0x0448
    GW_ACTIVATE_PRODUCTGROUP_NTF = 0x0449

    GW_GET_CONTACT_INPUT_LINK_LIST_REQ = 0x0460
    GW_GET_CONTACT_INPUT_LINK_LIST_CFM = 0x0461

    GW_SET_CONTACT_INPUT_LINK_REQ = 0x0462
    GW_SET_CONTACT_INPUT_LINK_CFM = 0x0463

    GW_REMOVE_CONTACT_INPUT_LINK_REQ = 0x0464
    GW_REMOVE_CONTACT_INPUT_LINK_CFM = 0x0465

    GW_GET_ACTIVATION_LOG_HEADER_REQ = 0x0500
    GW_GET_ACTIVATION_LOG_HEADER_CFM = 0x0501

    GW_CLEAR_ACTIVATION_LOG_REQ = 0x0502
    GW_CLEAR_ACTIVATION_LOG_CFM = 0x0503

    GW_GET_ACTIVATION_LOG_LINE_REQ = 0x0504
    GW_GET_ACTIVATION_LOG_LINE_CFM = 0x0505

    GW_ACTIVATION_LOG_UPDATED_NTF = 0x0506

    GW_GET_MULTIPLE_ACTIVATION_LOG_LINES_REQ = 0x0507
    GW_GET_MULTIPLE_ACTIVATION_LOG_LINES_NTF = 0x0508
    GW_GET_MULTIPLE_ACTIVATION_LOG_LINES_CFN = 0x0509

    GW_SET_UTC_REQ = 0x2000
    GW_SET_UTC_CFM = 0x2001

    GW_RTC_SET_TIME_ZONE_REQ = 0x2002
    GW_RTC_SET_TIME_ZONE_CFM = 0x2003

    GW_GET_LOCAL_TIME_REQ = 0x2004
    GW_GET_LOCAL_TIME_CFM = 0x2005

    GW_PASSWORD_ENTER_REQ = 0x3000
    GW_PASSWORD_ENTER_CFM = 0x3001

    GW_PASSWORD_CHANGE_REQ = 0x3002
    GW_PASSWORD_CHANGE_CFM = 0x3003
    GW_PASSWORD_CHANGED_NTF = 0x3004


class Originator(Enum):
    """Enum class for originator."""

    USER = 1
    RAIN = 2
    TIMER = 3
    UPS = 5  # UPC unit
    SAAC = 8  # Stand Alone Automatic Controls
    WIND = 9
    LOAD_SHEDDING = 11
    LOCAL_LIGHT = 12
    UNSPECIFIC_ENVIRONMENT_SENSOR = 13
    EMERGENCY = 255


class Priority(Enum):
    """Enum class for priority."""

    PROTECTION_HUMAN = 0
    PROTECTION_ENVIRONMENT = 1
    USER_LEVEL_1 = 2
    USER_LEVEL_2 = 3
    COMFORT_LEVEL_1 = 4
    COMFORT_LEVEL_2 = 5
    COMFORT_LEVEL_3 = 6
    COMFORT_LEVEL_4 = 7


class Velocity(Enum):
    """Enum class for velocity."""

    DEFAULT = 0
    SILENT = 1
    FAST = 2
    NOT_AVAILABLE = 255


class NodeTypeWithSubtype(Enum):
    """Enum class for node type plus sub type combined values."""

    # pylint: disable=invalid-name
    NO_TYPE = 0
    INTERIOR_VENETIAN_BLIND = 0x0040
    ROLLER_SHUTTER = 0x0080
    ADJUSTABLE_SLUTS_ROLLING_SHUTTER = 0x0081
    ADJUSTABLE_SLUTS_ROLLING_SHUTTER_WITH_PROJECTION = 0x0082
    VERTICAL_EXTERIOR_AWNING = 0x00C0
    WINDOW_OPENER = 0x0100
    WINDOW_OPENER_WITH_RAIN_SENSOR = 0x0101
    GARAGE_DOOR_OPENER = 0x0140
    LINAR_ANGULAR_POSITION_OF_GARAGE_DOOR = 0x140
    LIGHT = 0x0180
    LIGHT_ON_OFF = 0x01BA
    GATE_OPENER = 0x01C0
    GATE_OPENER_ANGULAR_POSITION = 0x01FA
    DOOR_LOCK = 0x0240
    WINDOW_LOCK = 0x0241
    VERTICAL_INTERIOR_BLINDS = 0x0280
    DUAL_ROLLER_SHUTTER = 0x0340
    ON_OFF_SWITCH = 0x03C0
    HORIZONTAL_AWNING = 0x0400
    EXTERIOR_VENETIAN_BLIND = 0x0440
    LOUVER_BLIND = 0x0480
    CURTAIN_TRACK = 0x04C0
    VENTILATION_POINT = 0x0500
    VENTILATION_POINT_AIR_INLET = 0x0502
    VENTILATION_POINT_AIR_TRANSFER = 0x0503
    VENTILATION_POINT_AIR_OUTLET = 0x0503
    EXTERIOR_HEATING = 0x0540
    SWINGING_SHUTTERS = 0x0600
    SWINGING_SHUTTER_WITH_INDEPENDENT_LEAVES = 0x0601


class NodeType(Enum):
    """Enum class for node types."""

    NO_TYPE = 0
    VENETIAN_BLIND = 1
    ROLLER_SHUTTER = 2
    AWNING = 3
    WINDOW_OPENER = 4
    GARAGE_OPENER = 5
    LIGHT = 6
    GATE_OPENER = 7
    ROLLING_DOOR_OPENER = 8
    LOCK = 9
    BLIND = 10
    SECURE_CONFIGURATION_DEVICE = 11
    BEACON = 12
    DUAL_SHUTTER = 13
    HEATING_TEMPERATURE_INTERFACE = 14
    ON_OFF_SWITCH = 15
    HORIZONTAL_AWNING = 16
    EXTERNAL_VENETIAN_BLIND = 17
    LOUVRE_BLINT = 18
    CURTAIN_TRACK = 19
    VENTILATION_POINT = 20
    EXTERIOR_HEATING = 21
    HEAT_PUMP = 22
    INTRUSION_ALARM = 23
    SWINGING_SHUTTER = 24


class NodeVariation(Enum):
    """Enum class for node variations."""

    NOT_SET = 0
    TOPHUNG = 1
    KIP = 2
    FLAT_ROOT = 3
    SKY_LIGHT = 3



"""Module for get version frame classes."""
from pyvlx.const import Command

from .frame import FrameBase


class FrameGetProtocolVersionRequest(FrameBase):
    """Frame for requesting protocol version."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_PROTOCOL_VERSION_REQ)


class FrameGetProtocolVersionConfirmation(FrameBase):
    """Frame for response for get protocol version requests."""

    PAYLOAD_LEN = 4

    def __init__(self, major_version=0, minor_version=0):
        """Init Frame."""
        super().__init__(Command.GW_GET_PROTOCOL_VERSION_CFM)
        self.major_version = major_version
        self.minor_version = minor_version

    @property
    def version(self):
        """Return formatted protocol version."""
        return "{}.{}".format(self.major_version, self.minor_version)

    def get_payload(self):
        """Return Payload."""
        return bytes(
            [self.major_version >> 8 & 255, self.major_version & 255,
             self.minor_version >> 8 & 255, self.minor_version & 255])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.major_version = payload[0] * 256 + payload[1]
        self.minor_version = payload[2] * 256 + payload[3]

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetProtocolVersionConfirmation version="{}"/>'.format(self.version)



"""Module for configuration."""
import yaml

from .exception import PyVLXException
from .log import PYVLXLOG


class Config:
    """Object for configuration."""

    DEFAULT_PORT = 51200

    def __init__(self, pyvlx, path=None, host=None, password=None, port=None):
        """Initialize Config class."""
        self.pyvlx = pyvlx
        self.host = host
        self.password = password
        self.port = port or self.DEFAULT_PORT
        if path is not None:
            self.read_config(path)

    def read_config(self, path):
        """Read configuration file."""
        PYVLXLOG.info('Reading config file: %s', path)
        try:
            with open(path, 'r') as filehandle:
                doc = yaml.load(filehandle)
                self.test_configuration(doc, path)
                self.host = doc['config']['host']
                self.password = doc['config']['password']
                if 'port' in doc['config']:
                    self.port = doc['config']['port']
        except FileNotFoundError as ex:
            raise PyVLXException('file does not exist: {0}'.format(ex))

    @staticmethod
    def test_configuration(doc, path):
        """Test if configuration file is sane."""
        if 'config' not in doc:
            raise PyVLXException('no element config found in: {0}'.format(path))
        if 'host' not in doc['config']:
            raise PyVLXException('no element host found in: {0}'.format(path))
        if 'password' not in doc['config']:
            raise PyVLXException('no element password found in: {0}'.format(path))



"""Module for Position class."""
from .exception import PyVLXException


class Position():
    """Class for storing a position."""

    UNKNOWN_POSITION = b'\xF7\xFF'
    MAX = 51200

    def __init__(self, raw=None, position=None, position_percent=None):
        """Initialize Position class."""
        if raw is not None:
            self.raw = self.from_raw(raw)
        elif position is not None:
            self.position = position
        elif position_percent is not None:
            self.position_percent = position_percent
        else:
            self.raw = Position.UNKNOWN_POSITION

    def __bytes__(self):
        """Convert object in byte representation."""
        return self.raw

    @property
    def known(self):
        """Known property, true if position is not in an unknown position."""
        return self.raw != Position.UNKNOWN_POSITION

    @property
    def open(self):
        """Return true if position is set to fully open."""
        return self.raw == b'\x00\x00'

    @property
    def closed(self):
        """Return true if position is set to fully closed."""
        return self.raw == bytes([self.MAX >> 8 & 255, self.MAX & 255])

    @property
    def position(self):
        """Position property."""
        return self.to_int(self.raw)

    @position.setter
    def position(self, position):
        """Setter of internal raw via position."""
        self.raw = self.from_int(position)

    @property
    def position_percent(self):
        """Position percent property."""
        # inclear why it returns a <property object> here
        return int(self.to_percent(self.raw))

    @position_percent.setter
    def position_percent(self, position_percent):
        """Setter of internal raw via percent position."""
        self.raw = self.from_percent(position_percent)

    @staticmethod
    def from_int(position):
        """Create raw out of position vlaue."""
        if not isinstance(position, int):
            raise PyVLXException("Position::position_has_to_be_int")
        if position < 0:
            raise PyVLXException("Position::position_has_to_be_positive")
        if position > Position.MAX:
            raise PyVLXException("Position::position_out_of_range")
        return bytes([position >> 8 & 255, position & 255])

    @staticmethod
    def to_int(raw):
        """Create int position value out of raw."""
        return raw[0] * 256 + raw[1]

    @staticmethod
    def from_percent(position_percent):
        """Create raw value out of percent position."""
        if not isinstance(position_percent, int):
            raise PyVLXException("Position::position_percent_has_to_be_int")
        if position_percent < 0:
            raise PyVLXException("Position::position_percent_has_to_be_positive")
        if position_percent > 100:
            raise PyVLXException("Position::position_percent_out_of_range")
        return bytes([position_percent*2, 0])

    @staticmethod
    def to_percent(raw):
        """Create percent position value out of raw."""
        # The first byte has the vlue from 0 to 200. Ignoring the second one.
        return int(raw[0]/2)

    @staticmethod
    def from_raw(raw):
        """Test if raw packets are valid for initialization of Position."""
        if not isinstance(raw, bytes):
            raise PyVLXException("Position::raw_must_be_bytes")
        if len(raw) != 2:
            raise PyVLXException("Position::raw_must_be_two_bytes")
        if raw != Position.UNKNOWN_POSITION and Position.to_int(raw) > Position.MAX:
            raise PyVLXException("position::raw_exceed_limit", raw=raw)
        return raw

    def __str__(self):
        """Return string representation of object."""
        if self.raw == Position.UNKNOWN_POSITION:
            return "UNKNOWN"
        return "{} %".format(self.position_percent)

    def __eq__(self, other):
        """Equal operator."""
        return self.raw == other.raw



"""Module for retrieving scene list from API."""
from .api_event import ApiEvent
from .frames import (FrameGetSceneListConfirmation,
                     FrameGetSceneListNotification, FrameGetSceneListRequest)
from .log import PYVLXLOG


class GetSceneList(ApiEvent):
    """Class for retrieving scene list from API."""

    def __init__(self, pyvlx):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False
        self.count_scenes = None
        self.scenes = []

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameGetSceneListConfirmation):
            self.count_scenes = frame.count_scenes
            # We are still waiting for FrameGetSceneListNotification(s)
            return False
        if isinstance(frame, FrameGetSceneListNotification):
            self.scenes.extend(frame.scenes)
            if frame.remaining_scenes != 0:
                # We are still waiting for FrameGetSceneListConfirmation(s)
                return False
            if self.count_scenes != len(self.scenes):
                PYVLXLOG.warning("Warning: number of received scenes does not match expected number")
            self.success = True
            return True
        return False

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetSceneListRequest()



"""Module for storing nodes."""
from .exception import PyVLXException
from .get_all_nodes_information import GetAllNodesInformation
from .get_node_information import GetNodeInformation
from .node import Node
from .node_helper import convert_frame_to_node


class Nodes:
    """Object for storing node objects."""

    def __init__(self, pyvlx):
        """Initialize Nodes object."""
        self.pyvlx = pyvlx
        self.__nodes = []

    def __iter__(self):
        """Iterator."""
        yield from self.__nodes

    def __getitem__(self, key):
        """Return node by name or by index."""
        if isinstance(key, int):
            return self.__nodes[key]
        for node in self.__nodes:
            if node.name == key:
                return node
        raise KeyError

    def __len__(self):
        """Return number of nodes."""
        return len(self.__nodes)

    def add(self, node):
        """Add Node, replace existing node if node with node_id is present."""
        if not isinstance(node, Node):
            raise TypeError()
        for i, j in enumerate(self.__nodes):
            if j.node_id == node.node_id:
                self.__nodes[i] = node
                return
        self.__nodes.append(node)

    def clear(self):
        """Clear internal node array."""
        self.__nodes = []

    async def load(self, node_id=None):
        """Load nodes from KLF 200, if no node_id is specified all nodes are loaded."""
        if node_id is not None:
            await self._load_node(node_id=node_id)
        else:
            await self._load_all_nodes()

    async def _load_node(self, node_id):
        """Load single node via API."""
        get_node_information = GetNodeInformation(pyvlx=self.pyvlx, node_id=node_id)
        await get_node_information.do_api_call()
        if not get_node_information.success:
            raise PyVLXException("Unable to retrieve node information")
        notification_frame = get_node_information.notification_frame
        node = convert_frame_to_node(self.pyvlx, notification_frame)
        self.add(node)

    async def _load_all_nodes(self):
        """Load all nodes via API."""
        get_all_nodes_information = GetAllNodesInformation(pyvlx=self.pyvlx)
        await get_all_nodes_information.do_api_call()
        if not get_all_nodes_information.success:
            raise PyVLXException("Unable to retrieve node information")
        self.clear()
        for notification_frame in get_all_nodes_information.notification_frames:
            node = convert_frame_to_node(self.pyvlx, notification_frame)
            self.add(node)



"""Module for storing and accessing scene list."""
from .exception import PyVLXException
from .get_scene_list import GetSceneList
from .scene import Scene


class Scenes():
    """Class for storing and accessing ."""

    def __init__(self, pyvlx):
        """Initialize Scenes class."""
        self.pyvlx = pyvlx
        self.__scenes = []

    def __iter__(self):
        """Iterator."""
        yield from self.__scenes

    def __getitem__(self, key):
        """Return scene by name or by index."""
        if isinstance(key, int):
            return self.__scenes[key]
        for scene in self.__scenes:
            if scene.name == key:
                return scene
        raise KeyError

    def __len__(self):
        """Return number of scenes."""
        return len(self.__scenes)

    def add(self, scene):
        """Add scene, replace existing scene if scene with scene_id is present."""
        if not isinstance(scene, Scene):
            raise TypeError()
        for i, j in enumerate(self.__scenes):
            if j.scene_id == scene.scene_id:
                self.__scenes[i] = scene
                return
        self.__scenes.append(scene)

    def clear(self):
        """Clear internal scenes array."""
        self.__scenes = []

    async def load(self):
        """Load scenes from KLF 200."""
        get_scene_list = GetSceneList(pyvlx=self.pyvlx)
        await get_scene_list.do_api_call()
        if not get_scene_list.success:
            raise PyVLXException("Unable to retrieve scene information")
        for scene in get_scene_list.scenes:
            self.add(Scene(pyvlx=self.pyvlx, scene_id=scene[0], name=scene[1]))



"""Module for window openers."""
from .command_send import CommandSend
from .exception import PyVLXException
from .node import Node
from .position import Position


class OpeningDevice(Node):
    """Meta class for opening device with one main parameter for position."""

    def __init__(self, pyvlx, node_id, name):
        """Initialize opening device."""
        super().__init__(pyvlx=pyvlx, node_id=node_id, name=name)
        self.position = Position()

    async def set_position(self, position):
        """Set window to desired position."""
        command_send = CommandSend(pyvlx=self.pyvlx, node_id=self.node_id, position=position)
        await command_send.do_api_call()
        if not command_send.success:
            raise PyVLXException("Unable to send command")
        self.position = position
        await self.after_update()

    async def open(self):
        """Open window."""
        await self.set_position(Position(position_percent=0))

    async def close(self):
        """Close window."""
        await self.set_position(Position(position_percent=100))


class Window(OpeningDevice):
    """Window object."""

    def __init__(self, pyvlx, node_id, name, rain_sensor=False):
        """Initialize Window class."""
        super().__init__(pyvlx=pyvlx, node_id=node_id, name=name)
        self.rain_sensor = rain_sensor

    def __str__(self):
        """Return object as readable string."""
        return '<{} name="{}" ' \
            'node_id="{}" rain_sensor={}/>' \
            .format(
                type(self).__name__,
                self.name,
                self.node_id, self.rain_sensor)


class Blind(OpeningDevice):
    """Blind objects."""


class RollerShutter(OpeningDevice):
    """RollerShutter object."""



"""Module for Serial Line Internet Protocol (SLIP)."""

SLIP_END = 0xC0
SLIP_ESC = 0xDB
SLIP_ESC_END = 0xDC
SLIP_ESC_ESC = 0xDD


def is_slip(raw):
    """Check if raw is a SLIP packet."""
    if len(raw) < 2:
        return False
    return raw[0] == SLIP_END and SLIP_END in raw[1:]


def decode(raw):
    """Decode SLIP message."""
    return raw \
        .replace(bytes([SLIP_ESC, SLIP_ESC_END]), bytes([SLIP_END])) \
        .replace(bytes([SLIP_ESC, SLIP_ESC_ESC]), bytes([SLIP_ESC]))


def encode(raw):
    """Encode SLIP message."""
    return raw \
        .replace(bytes([SLIP_ESC]), bytes([SLIP_ESC, SLIP_ESC_ESC])) \
        .replace(bytes([SLIP_END]), bytes([SLIP_ESC, SLIP_ESC_END]))


def get_next_slip(raw):
    """
    Get the next slip packet from raw data.

    Returns the extracted packet plus the raw data with the remaining data stream.
    """
    if not is_slip(raw):
        return None, raw
    length = raw[1:].index(SLIP_END)
    slip_packet = decode(raw[1:length+1])
    new_raw = raw[length+2:]
    return slip_packet, new_raw


def slip_pack(raw):
    """Pack raw message to complete slip message."""
    return bytes([SLIP_END]) + encode(raw) + bytes([SLIP_END])



"""Module for retrieving node information from API."""
from .api_event import ApiEvent
from .frames import (FrameGetNodeInformationConfirmation,
                     FrameGetNodeInformationNotification,
                     FrameGetNodeInformationRequest)


class GetNodeInformation(ApiEvent):
    """Class for retrieving node informationfrom API."""

    def __init__(self, pyvlx, node_id):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx)
        self.node_id = node_id
        self.success = False
        self.notification_frame = None

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameGetNodeInformationConfirmation) and frame.node_id == self.node_id:
            # We are still waiting for GetNodeInformationNotification
            return False
        if isinstance(frame, FrameGetNodeInformationNotification) and frame.node_id == self.node_id:
            self.notification_frame = frame
            self.success = True
            return True
        return False

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetNodeInformationRequest(node_id=self.node_id)



"""Module for requesting change of node name."""
from pyvlx.const import Command, NodeVariation
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase


class FrameNodeInformationChangedNotification(FrameBase):
    """Frame for notification for set node name."""

    PAYLOAD_LEN = 69

    def __init__(self, node_id=0, name=None, order=0, placement=0, node_variation=NodeVariation.NOT_SET):
        """Init Frame."""
        super().__init__(Command.GW_NODE_INFORMATION_CHANGED_NTF)
        self.node_id = node_id
        self.name = name
        self.order = order
        self.placement = placement
        self.node_variation = node_variation

    def get_payload(self):
        """Return Payload."""
        payload = bytes([self.node_id])
        payload += string_to_bytes(self.name, 64)
        payload += bytes([self.order >> 8 & 255, self.order & 255])
        payload += bytes([self.placement])
        payload += bytes([self.node_variation.value])
        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_id = payload[0]
        self.name = bytes_to_string(payload[1:65])
        self.order = payload[65] * 256 + payload[66]
        self.placement = payload[67]
        self.node_variation = NodeVariation(payload[68])

    def __str__(self):
        """Return human readable string."""
        return '<FrameNodeInformationChangedNotification node_id={} name="{}" order={} ' \
            'placement={} node_variation="{}"/>'.format(
                self.node_id, self.name, self.order,
                self.placement, self.node_variation)



"""Module for accessing KLF 200 gateway with python."""

# flake8: noqa
from .pyvlx import PyVLX
from .exception import PyVLXException
from .nodes import Nodes
from .position import Position
from .opening_device import OpeningDevice, Window, RollerShutter, Blind
from .scenes import Scenes
from .scene import Scene



"""Base class for waiting for a specific answer frame from velux ap.."""
import asyncio


class ApiEvent():
    """Base class for waiting a specific frame from API connection."""

    def __init__(self, pyvlx, timeout_in_seconds=10):
        """Initialize ApiEvent."""
        self.pyvlx = pyvlx
        self.response_received_or_timeout = asyncio.Event()

        self.success = False
        self.timeout_in_seconds = timeout_in_seconds
        self.timeout_callback = None
        self.timeout_handle = None

    async def do_api_call(self):
        """Start. Sending and waiting for answer."""
        self.pyvlx.connection.register_frame_received_cb(
            self.response_rec_callback)
        await self.send_frame()
        await self.start_timeout()
        await self.response_received_or_timeout.wait()
        await self.stop_timeout()
        self.pyvlx.connection.unregister_frame_received_cb(self.response_rec_callback)

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        raise NotImplementedError('handle_frame has to be implemented')

    async def send_frame(self):
        """Send frame to API connection."""
        await self.pyvlx.send_frame(self.request_frame())

    def request_frame(self):
        """Construct initiating framw."""
        raise NotImplementedError('send_frame has to be implemented')

    async def response_rec_callback(self, frame):
        """Handle frame. Callback from internal api connection."""
        if await self.handle_frame(frame):
            self.response_received_or_timeout.set()

    def timeout(self):
        """Handle timeout for not having received expected frame."""
        self.response_received_or_timeout.set()

    async def start_timeout(self):
        """Start timeout."""
        self.timeout_handle = self.pyvlx.connection.loop.call_later(
            self.timeout_in_seconds, self.timeout)

    async def stop_timeout(self):
        """Stop timeout."""
        self.timeout_handle.cancel()



"""Helper module for Node objects."""
from .const import NodeTypeWithSubtype
from .log import PYVLXLOG
from .opening_device import Blind, RollerShutter, Window


def convert_frame_to_node(pyvlx, frame):
    """Convert FrameGet[All]Node[s]InformationNotification into Node object."""
    if frame.node_type == NodeTypeWithSubtype.WINDOW_OPENER:
        return Window(pyvlx=pyvlx, node_id=frame.node_id, name=frame.name, rain_sensor=False)
    if frame.node_type == NodeTypeWithSubtype.WINDOW_OPENER_WITH_RAIN_SENSOR:
        return Window(pyvlx=pyvlx, node_id=frame.node_id, name=frame.name, rain_sensor=True)
    if frame.node_type == NodeTypeWithSubtype.ROLLER_SHUTTER:
        return RollerShutter(pyvlx=pyvlx, node_id=frame.node_id, name=frame.name)
    if frame.node_type == NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND or \
            frame.node_type == NodeTypeWithSubtype.VERTICAL_INTERIOR_BLINDS or \
            frame.node_type == NodeTypeWithSubtype.EXTERIOR_VENETIAN_BLIND or \
            frame.node_type == NodeTypeWithSubtype.LOUVER_BLIND:
        return Blind(pyvlx=pyvlx, node_id=frame.node_id, name=frame.name)

    PYVLXLOG.warning("%s not implemented", frame.node_type)
    return None



"""Module for get scene list frame classes."""
from pyvlx.const import Command
from pyvlx.exception import PyVLXException
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase


class FrameGetSceneListRequest(FrameBase):
    """Frame for get scene list request."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_SCENE_LIST_REQ)


class FrameGetSceneListConfirmation(FrameBase):
    """Frame for confirmation for scene list request."""

    PAYLOAD_LEN = 1

    def __init__(self, count_scenes=0):
        """Init Frame."""
        super().__init__(Command.GW_GET_SCENE_LIST_CFM)
        self.count_scenes = count_scenes

    def get_payload(self):
        """Return Payload."""
        return bytes([self.count_scenes])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.count_scenes = payload[0]

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetSceneListConfirmation count_scenes={}/>'.format(self.count_scenes)


class FrameGetSceneListNotification(FrameBase):
    """Frame for scene list notification."""

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_SCENE_LIST_NTF)
        self.scenes = []
        self.remaining_scenes = 0

    def get_payload(self):
        """Return Payload."""
        ret = bytes([len(self.scenes)])
        for number, name in self.scenes:
            ret += bytes([number])
            ret += string_to_bytes(name, 64)
        ret += bytes([self.remaining_scenes])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        number_of_objects = payload[0]
        self.remaining_scenes = payload[-1]
        predicted_len = number_of_objects * 65 + 2
        if len(payload) != predicted_len:
            raise PyVLXException('scene_list_notification_wrong_length')
        self.scenes = []
        for i in range(number_of_objects):
            scene = payload[(i*65+1):(i*65+66)]
            number = scene[0]
            name = bytes_to_string(scene[1:])
            self.scenes.append((number, name))

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetSceneListNotification scenes={} remaining_scenes={}>'.format(self.scenes, self.remaining_scenes)



"""Module for get node information from gateway."""
from enum import Enum
from datetime import datetime
import struct

from pyvlx.const import Command, NodeTypeWithSubtype, NodeVariation, Velocity
from pyvlx.position import Position
from pyvlx.alias_array import AliasArray
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase


class FrameGetNodeInformationRequest(FrameBase):
    """Frame for get node information request."""

    PAYLOAD_LEN = 1

    def __init__(self, node_id=None):
        """Init Frame."""
        super().__init__(Command.GW_GET_NODE_INFORMATION_REQ)
        self.node_id = node_id

    def get_payload(self):
        """Return Payload."""
        return bytes([self.node_id])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_id = payload[0]

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetNodeInformationRequest node_id={}/>'.format(self.node_id)


class NodeInformationStatus(Enum):
    """Enum for node information status."""

    OK = 0  # pylint: disable=invalid-name
    Error_Request_Rejected = 1
    Error_Invalid_Node_Index = 2


class FrameGetNodeInformationConfirmation(FrameBase):
    """Frame for confirmation for node information request."""

    PAYLOAD_LEN = 2

    def __init__(self, status=NodeInformationStatus.OK, node_id=None):
        """Init Frame."""
        super().__init__(Command.GW_GET_NODE_INFORMATION_CFM)
        self.status = status
        self.node_id = node_id

    def get_payload(self):
        """Return Payload."""
        return bytes([self.status.value, self.node_id])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = NodeInformationStatus(payload[0])
        self.node_id = payload[1]

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetNodeInformationConfirmation node_id={} status=\'{}\'/>'.format(self.node_id, self.status)


class FrameGetNodeInformationNotification(FrameBase):
    """Frame for notification of note information request."""

    PAYLOAD_LEN = 124

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_NODE_INFORMATION_NTF)
        self.node_id = 0
        self.order = 0
        self.placement = 0
        self.name = ""
        self.velocity = Velocity.DEFAULT
        self.node_type = NodeTypeWithSubtype.NO_TYPE
        self.product_group = 0
        self.product_type = 0
        self.node_variation = NodeVariation.NOT_SET
        self.power_mode = 0
        self.build_number = 0
        self._serial_number = bytes(8)
        self.state = 0
        self.current_position = Position()
        self.target = Position()
        self.current_position_fp1 = Position()
        self.current_position_fp2 = Position()
        self.current_position_fp3 = Position()
        self.current_position_fp4 = Position()
        self.remaining_time = 0
        self.timestamp = 0
        self.alias_array = AliasArray()

    @property
    def serial_number(self):
        """Property for serial number in a human readable way."""
        return ":".join("{:02x}".format(c) for c in self._serial_number)

    def get_payload(self):
        """Return Payload."""
        payload = bytes()
        payload += bytes([self.node_id])
        payload += bytes([self.order >> 8 & 255, self.order & 255])
        payload += bytes([self.placement])
        payload += bytes(string_to_bytes(self.name, 64))
        payload += bytes([self.velocity.value])
        payload += bytes([self.node_type.value >> 8 & 255, self.node_type.value & 255])
        payload += bytes([self.product_group])
        payload += bytes([self.product_type])
        payload += bytes([self.node_variation.value])
        payload += bytes([self.power_mode])
        payload += bytes([self.build_number])  # <-- hey @VELUX: your documentation is wrong here
        payload += bytes(self._serial_number)
        payload += bytes([self.state])
        payload += bytes(self.current_position.raw)
        payload += bytes(self.target.raw)
        payload += bytes(self.current_position_fp1.raw)
        payload += bytes(self.current_position_fp2.raw)
        payload += bytes(self.current_position_fp3.raw)
        payload += bytes(self.current_position_fp4.raw)
        payload += bytes([self.remaining_time >> 8 & 255, self.remaining_time & 255])
        payload += struct.pack(">I", self.timestamp)
        payload += bytes(self.alias_array)
        return payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_id = payload[0]
        self.order = payload[1] * 256 + payload[2]
        self.placement = payload[3]
        self.name = bytes_to_string(payload[4:68])
        self.velocity = Velocity(payload[68])
        self.node_type = NodeTypeWithSubtype(payload[69] * 256 + payload[70])
        self.product_group = payload[71]
        self.product_type = payload[72]
        self.node_variation = NodeVariation(payload[73])
        self.power_mode = payload[74]
        self.build_number = payload[75]  # <-- hey @VELUX: your documentation is wrong here
        self._serial_number = payload[76:84]
        self.state = payload[84]
        self.current_position = Position(payload[85:87])
        self.target = Position(payload[87:89])
        self.current_position_fp1 = Position(payload[89:91])
        self.current_position_fp2 = Position(payload[91:93])
        self.current_position_fp3 = Position(payload[93:95])
        self.current_position_fp4 = Position(payload[95:97])
        self.remaining_time = payload[97] * 256 + payload[98]
        self.timestamp = struct.unpack(">I", payload[99:103])[0]
        self.alias_array = AliasArray(payload[103:125])

    @property
    def timestamp_formatted(self):
        """Return time as human readable string."""
        return datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S')

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetNodeInformationNotification node_id={} order={} ' \
            'placement={} name=\'{}\' velocity={} node_type=\'{}\' product_group={} ' \
            'product_type={} node_variation={} power_mode={} build_number={} ' \
            'serial_number=\'{}\' state={} current_position=\'{}\' ' \
            'target=\'{}\' current_position_fp1=\'{}\' current_position_fp2=\'{}\' ' \
            'current_position_fp3=\'{}\' current_position_fp4=\'{}\' ' \
            'remaining_time={} time=\'{}\' alias_array=\'{}\'/>'.format(
                self.node_id,
                self.order,
                self.placement,
                self.name,
                self.velocity,
                self.node_type,
                self.product_group,
                self.product_type,
                self.node_variation,
                self.power_mode,
                self.build_number,
                self.serial_number,
                self.state,
                self.current_position,
                self.target,
                self.current_position_fp1,
                self.current_position_fp2,
                self.current_position_fp3,
                self.current_position_fp4,
                self.remaining_time,
                self.timestamp_formatted,
                self.alias_array)



"""Module for discover nodes requests."""
from pyvlx.const import Command, NodeType

from .frame import FrameBase


class FrameDiscoverNodesRequest(FrameBase):
    """Frame for discover nodes request."""

    PAYLOAD_LEN = 1

    def __init__(self, node_type=NodeType.NO_TYPE):
        """Init Frame."""
        super().__init__(Command.GW_CS_DISCOVER_NODES_REQ)
        self.node_type = node_type

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.node_type.value])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_type = NodeType(payload[0])

    def __str__(self):
        """Return human readable string."""
        return '<FrameDiscoverNodesRequest node_type=\'{}\'/>'.format(self.node_type)


class FrameDiscoverNodesConfirmation(FrameBase):
    """Frame for discover nodes confirmation."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_CS_DISCOVER_NODES_CFM)


class FrameDiscoverNodesNotification(FrameBase):
    """Frame for discover nodes notification."""

    PAYLOAD_LEN = 131

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_CS_DISCOVER_NODES_NTF)
        self.payload = b'\0' * 131

    def get_payload(self):
        """Return Payload."""
        return self.payload

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.payload = payload

    def __str__(self):
        """Return human readable string."""
        return '<FrameDiscoverNodesNotification payload=\'{}\'/>'.format(":".join("{:02x}".format(c) for c in self.payload))



"""Module for retrieving node information from API."""
from .api_event import ApiEvent
from .frames import (FrameGetAllNodesInformationConfirmation,
                     FrameGetAllNodesInformationFinishedNotification,
                     FrameGetAllNodesInformationNotification,
                     FrameGetAllNodesInformationRequest)
from .log import PYVLXLOG


class GetAllNodesInformation(ApiEvent):
    """Class for retrieving node informationfrom API."""

    def __init__(self, pyvlx):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx)
        self.number_of_nodes = 0
        self.success = False
        self.notification_frames = []

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameGetAllNodesInformationConfirmation):
            self.number_of_nodes = frame.number_of_nodes
            # We are still waiting for FrameGetAllNodesInformationNotification
            return False
        if isinstance(frame, FrameGetAllNodesInformationNotification):
            self.notification_frames.append(frame)
        if isinstance(frame, FrameGetAllNodesInformationFinishedNotification):
            if self.number_of_nodes != len(self.notification_frames):
                PYVLXLOG.warning("Number of received scenes does not match expected number")
            self.success = True
            return True
        return False

    def request_frame(self):
        """Construct initiating frame."""
        return FrameGetAllNodesInformationRequest()



"""Module for updating nodes via frames."""
from .frames import FrameNodeStatePositionChangedNotification
from .opening_device import OpeningDevice


async def update_nodes(self, frame):
    """Update nodes via frame, usually received by house monitor."""
    if isinstance(frame, FrameNodeStatePositionChangedNotification):
        node = self.nodes[frame.node_id]
        if isinstance(node, OpeningDevice):
            node.position = frame.current_position
            await node.after_update()



"""Module for sending command to gw."""
from enum import Enum

from pyvlx.const import Command, Originator, Priority, Velocity

from .frame import FrameBase


class FrameActivateSceneRequest(FrameBase):
    """Frame for sending command to gw."""

    PAYLOAD_LEN = 6

    def __init__(self, scene_id=None, session_id=None, originator=Originator.USER, velocity=Velocity.DEFAULT):
        """Init Frame."""
        super().__init__(Command.GW_ACTIVATE_SCENE_REQ)
        self.scene_id = scene_id
        self.session_id = session_id
        self.originator = originator
        self.priority = Priority.USER_LEVEL_2
        self.velocity = velocity

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.session_id >> 8 & 255, self.session_id & 255])
        ret += bytes([self.originator.value])
        ret += bytes([self.priority.value])
        ret += bytes([self.scene_id])
        ret += bytes([self.velocity.value])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.session_id = payload[0]*256 + payload[1]
        self.originator = Originator(payload[2])
        self.priority = Priority(payload[3])
        self.scene_id = payload[4]
        self.velocity = Velocity(payload[5])

    def __str__(self):
        """Return human readable string."""
        return '<FrameActivateSceneRequest scene_id={} session_id={} originator={} velocity={}/>'.format(
            self.scene_id, self.session_id, self.originator, self.velocity)


class ActivateSceneConfirmationStatus(Enum):
    """Enum class for status of command send confirmation."""

    ACCEPTED = 0
    ERROR_INVALID_PARAMETER = 1
    ERROR_REQUEST_REJECTED = 2


class FrameActivateSceneConfirmation(FrameBase):
    """Frame for confirmation of command send frame."""

    PAYLOAD_LEN = 3

    def __init__(self, session_id=None, status=None):
        """Init Frame."""
        super().__init__(Command.GW_ACTIVATE_SCENE_CFM)
        self.session_id = session_id
        self.status = status

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.status.value])
        ret += bytes([self.session_id >> 8 & 255, self.session_id & 255])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = ActivateSceneConfirmationStatus(payload[0])
        self.session_id = payload[1]*256 + payload[2]

    def __str__(self):
        """Return human readable string."""
        return '<FrameActivateSceneConfirmation session_id={} status={}/>'.format(self.session_id, self.status)



"""Module for handling the login to API."""
from .api_event import ApiEvent
from .frames import (FramePasswordEnterConfirmation, FramePasswordEnterRequest,
                     PasswordEnterConfirmationStatus)


class Login(ApiEvent):
    """Class for handling login to API."""

    def __init__(self, pyvlx, password):
        """Initialize login class."""
        super().__init__(pyvlx=pyvlx)
        self.password = password
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FramePasswordEnterConfirmation):
            return False
        if frame.status == PasswordEnterConfirmationStatus.FAILED:
            self.success = False
        if frame.status == PasswordEnterConfirmationStatus.SUCCESSFUL:
            self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FramePasswordEnterRequest(password=self.password)



"""Module for setting up PyVLX pypi object."""
import io
import os

from setuptools import setup, find_packages

REQUIRES = [
    'PyYAML'
]

PKG_ROOT = os.path.dirname(__file__)

VERSION = '0.2.4'

# Error-handling here is to allow package to be built w/o README included
try:
    README = io.open(os.path.join(PKG_ROOT, 'README.md'), encoding='utf-8').read()
except IOError:
    README = ''

setup(
    name='pyvlx',

    version=VERSION,
    download_url='https://github.com/Julius2342/pyvlx/archive/'+VERSION+'.zip',
    url='https://github.com/Julius2342/pyvlx',

    description="PyVLX is a wrapper for the Velux KLF 200 API.\n\nPyVLX enables you to run scenes and or open and close velux windows.",
    long_description=README,
    author='Julius Mittenzwei',
    author_email='julius@mittenzwei.com',
    license='LGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
        ],
    packages=find_packages(),
    install_requires=REQUIRES,
    keywords='velux KLF 200 home automation',
    zip_safe=False)



"""Module for requesting change of node name."""
from enum import Enum

from pyvlx.const import Command
from pyvlx.string_helper import bytes_to_string, string_to_bytes

from .frame import FrameBase


class FrameSetNodeNameRequest(FrameBase):
    """Frame for requesting node name change."""

    PAYLOAD_LEN = 65

    def __init__(self, node_id=0, name=None):
        """Init Frame."""
        super().__init__(Command.GW_SET_NODE_NAME_REQ)
        self.node_id = node_id
        self.name = name

    def get_payload(self):
        """Return Payload."""
        ret = bytes([self.node_id])
        ret += string_to_bytes(self.name, 64)
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.node_id = payload[0]
        self.name = bytes_to_string(payload[1:65])

    def __str__(self):
        """Return human readable string."""
        return '<FrameSetNodeNameRequest node_id={} name="{}"/>'.format(self.node_id, self.name)


class SetNodeNameConfirmationStatus(Enum):
    """Enum class for status of password enter confirmation."""

    OK = 0
    ERROR_REQUEST_REJECTED = 1
    ERROR_INVALID_SYSTEM_TABLE_INDEX = 2


class FrameSetNodeNameConfirmation(FrameBase):
    """Frame for confirmation for set node name."""

    PAYLOAD_LEN = 2

    def __init__(self, status=SetNodeNameConfirmationStatus.OK, node_id=0):
        """Init Frame."""
        super().__init__(Command.GW_SET_NODE_NAME_CFM)
        self.status = status
        self.node_id = node_id

    def get_payload(self):
        """Return Payload."""
        return bytes([self.status.value, self.node_id])

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.status = SetNodeNameConfirmationStatus(payload[0])
        self.node_id = payload[1]

    def __str__(self):
        """Return human readable string."""
        return '<FrameSetNodeNameConfirmation node_id={} status="{}"/>'.format(self.node_id, self.status)



"""Module for sending command to gw."""
import struct
from datetime import datetime

from pyvlx.const import Command

from .frame import FrameBase


class FrameSetUTCRequest(FrameBase):
    """Frame for sending command to gw."""

    PAYLOAD_LEN = 4

    def __init__(self, timestamp=0):
        """Init Frame."""
        super().__init__(Command.GW_SET_UTC_REQ)
        self.timestamp = timestamp

    @property
    def timestamp_formatted(self):
        """Return time as human readable string."""
        return datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S')

    def get_payload(self):
        """Return Payload."""
        return struct.pack(">I", self.timestamp)

    def from_payload(self, payload):
        """Init frame from binary data."""
        self.timestamp = struct.unpack(">I", payload[0:4])[0]

    def __str__(self):
        """Return human readable string."""
        return '<FrameSetUTCRequest time="{}"/>'.format(self.timestamp_formatted)



"""Module for changing a node name."""
from .api_event import ApiEvent
from .frames import (FrameSetNodeNameConfirmation, FrameSetNodeNameRequest,
                     SetNodeNameConfirmationStatus)


class SetNodeName(ApiEvent):
    """Class for changing the name of a node via API."""

    def __init__(self, pyvlx, node_id, name):
        """Initialize class."""
        super().__init__(pyvlx=pyvlx)
        self.node_id = node_id
        self.name = name
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameSetNodeNameConfirmation):
            return False
        self.success = frame.status == SetNodeNameConfirmationStatus.OK
        return True

    def request_frame(self):
        """Construct initiating frame."""
        return FrameSetNodeNameRequest(node_id=self.node_id, name=self.name)



"""Module for setting UTC time within gateway."""
import time
from .api_event import ApiEvent
from .exception import PyVLXException
from .frames import (FrameSetUTCConfirmation,
                     FrameSetUTCRequest)


class SetUTC(ApiEvent):
    """Class for setting UTC time within gateway."""

    def __init__(self, pyvlx):
        """Initialize login class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False

    async def handle_frame(self, frame):
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameSetUTCConfirmation):
            return False
        self.success = True
        return True

    def request_frame(self):
        """Construct initiating frame."""
        timestamp = int(time.time())
        return FrameSetUTCRequest(timestamp=timestamp)


async def set_utc(pyvlx):
    """Enable house status monitor."""
    setutc = SetUTC(pyvlx=pyvlx)
    await setutc.do_api_call()
    if not setutc.success:
        raise PyVLXException("Unable to set utc.")



"""Module for get version frame classes."""
from pyvlx.const import Command

from .frame import FrameBase


class FrameGetVersionRequest(FrameBase):
    """Frame for requesting version."""

    PAYLOAD_LEN = 0

    def __init__(self):
        """Init Frame."""
        super().__init__(Command.GW_GET_VERSION_REQ)


class FrameGetVersionConfirmation(FrameBase):
    """Frame for response for get version requests."""

    PAYLOAD_LEN = 9

    def __init__(self, software_version=bytes(6), hardware_version=0):
        """Init Frame."""
        super().__init__(Command.GW_GET_VERSION_CFM)
        if isinstance(software_version, str):
            software_version = bytes(int(c) for c in software_version.split("."))
        self._software_version = software_version
        self.hardware_version = hardware_version
        self.product_group = 14
        self.product_type = 3

    @property
    def version(self):
        """Return formatted version."""
        return "{}: Software version: {}, hardware version: {}".format(
            self.product, self.software_version, self.hardware_version)

    @property
    def software_version(self):
        """Return software version as human readable string."""
        return '.'.join(str(c) for c in self._software_version)

    @property
    def product(self):
        """Return product as human readable string."""
        if self.product_group == 14 and self.product_type == 3:
            return "KLF 200"
        return "Unknown Product: {}:{}".format(self.product_group, self.product_type)

    def get_payload(self):
        """Return Payload."""
        ret = self._software_version
        ret += bytes([self.hardware_version, self.product_group, self.product_type])
        return ret

    def from_payload(self, payload):
        """Init frame from binary data."""
        self._software_version = payload[0:6]
        self.hardware_version = payload[6]
        self.product_group = payload[7]
        self.product_type = payload[8]

    def __str__(self):
        """Return human readable string."""
        return '<FrameGetVersionConfirmation software_version="{}" ' \
            'harware_version="{}" product="{}"/>'.format(
                self.software_version, self.hardware_version,
                self.product)



"""Module for PyVLX Exceptions."""


class PyVLXException(Exception):
    """Default PyVLX Exception."""

    def __init__(self, description, **kwargs):
        """Initialize PyVLXException class."""
        super().__init__(description)
        self.description = description
        self.parameter = kwargs

    def _format_parameter(self):
        return " ".join(['%s="%s"' % (key, value) for (key, value) in sorted(self.parameter.items())])

    def __str__(self):
        """Return object as readable string."""
        return '<PyVLXException description="{0}" {1}/>' \
            .format(self.description, self._format_parameter())



"""Module for scene."""
from .activate_scene import ActivateScene
from .exception import PyVLXException


class Scene:
    """Object for scene."""

    def __init__(self, pyvlx, scene_id, name):
        """Initialize Scene object."""
        self.pyvlx = pyvlx
        self.scene_id = scene_id
        self.name = name

    async def run(self):
        """Run scene."""
        activate_scene = ActivateScene(pyvlx=self.pyvlx, scene_id=self.scene_id)
        await activate_scene.do_api_call()
        if not activate_scene.success:
            raise PyVLXException("Unable to activate scene")

    def __str__(self):
        """Return object as readable string."""
        return '<Scene name="{0}" ' \
            'id="{1}" />' \
            .format(
                self.name,
                self.scene_id)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__
